//*********************************************************************************************
// KProcCheck Kernel Driver (Proof-of-Concept) 
// Version 0.1
// by Chew Keong TAN
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, and/or sell copies of the Software, and to permit persons
// to whom the Software is furnished to do so, provided that the above
// copyright notice(s) and this permission notice appear in all copies of
// the Software and that both the above copyright notice(s) and this
// permission notice appear in supporting documentation.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
// OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
// HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
// INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
// FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
// NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
// WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
//*********************************************************************************************

#include <ntddk.h>
#include <ndis.h>
#include <stdio.h>
#include <windef.h>

#define INTERRUPT_OFF _asm{cli}
#define INTERRUPT_ON  _asm{sti}
#define NT_PROCNAMELEN  16


//*********************************************************************************************
// Type definitions
//
//*********************************************************************************************
typedef struct _PROCINFO
{
    DWORD pId;
    char pName[NT_PROCNAMELEN];
} PROCINFO;

typedef struct _DRIVERINFO
{
	DWORD loadAddr;
	DWORD imageSize;
	char drvName[MAX_PATH];
} DRIVERINFO;

typedef struct _SDTINFO
{
	DWORD sdtAddr;
	DWORD isHooked;
} SDTINFO;

#pragma pack(1)
typedef struct ServiceDescriptorEntry {
	PDWORD ServiceTable;
	PDWORD CounterTableBase;
	DWORD  ServiceLimit;
	PBYTE  ArgumentTable;
} SDT;
#pragma pack()


#define DRV_NAME	KProcCheck

#define _DRV_LINK(_name)   \\DosDevices\\ ##_name
#define _DRV_DEVICE(_name) \\Device\\ ##_name
#define DRV_DEVICE         _DRV_DEVICE(DRV_NAME)
#define DRV_LINK           _DRV_LINK(DRV_NAME)

#define _CSTRING(_text) #_text
#define CSTRING(_text) _CSTRING (_text)

#define _USTRING(_text) L##_text
#define USTRING(_text) _USTRING (_text)


#define PRESET_UNICODE_STRING(_var,_buffer) \
        UNICODE_STRING _var = \
            { \
            sizeof (USTRING (_buffer)) - sizeof (WORD), \
            sizeof (USTRING (_buffer)), \
            USTRING (_buffer) \
            };


PRESET_UNICODE_STRING(gDeviceName,       CSTRING(DRV_DEVICE));
PRESET_UNICODE_STRING(gSymbolicLinkName, CSTRING(DRV_LINK));


//*********************************************************************************************
// ioctl codes
//
//*********************************************************************************************

#define IOCTL_CODE1 CTL_CODE(FILE_DEVICE_UNKNOWN, 0x800, METHOD_OUT_DIRECT, FILE_READ_ACCESS + FILE_WRITE_ACCESS)
#define IOCTL_CODE2 CTL_CODE(FILE_DEVICE_UNKNOWN, 0x801, METHOD_OUT_DIRECT, FILE_READ_ACCESS + FILE_WRITE_ACCESS)
#define IOCTL_CODE3 CTL_CODE(FILE_DEVICE_UNKNOWN, 0x802, METHOD_OUT_DIRECT, FILE_READ_ACCESS + FILE_WRITE_ACCESS)
#define IOCTL_CODE4 CTL_CODE(FILE_DEVICE_UNKNOWN, 0x803, METHOD_OUT_DIRECT, FILE_READ_ACCESS + FILE_WRITE_ACCESS)
#define IOCTL_CODE5 CTL_CODE(FILE_DEVICE_UNKNOWN, 0x804, METHOD_OUT_DIRECT, FILE_READ_ACCESS + FILE_WRITE_ACCESS)
#define IOCTL_CODE6 CTL_CODE(FILE_DEVICE_UNKNOWN, 0x805, METHOD_BUFFERED, FILE_READ_ACCESS + FILE_WRITE_ACCESS)



//*********************************************************************************************
// Globals
//
//*********************************************************************************************

PDEVICE_OBJECT gpDeviceObject = NULL;
DWORD gPsLoadedModuleList = 0;
DWORD gKiWaitInListHead = 0;
DWORD gKiWaitOutListHead = 0;
DWORD gKiDispatcherReadyListHead = 0;
DWORD gProcessNameOffset = 0;
SDT *KeServiceDescriptorTableShadow = NULL;


//*********************************************************************************************
// Offsets for Win2k
//
//*********************************************************************************************

#define ACTIVELINK_OFFSETFROM_EPROC			(0xa0)
#define PROCNAME_OFFSETFROM_ACTIVELINK		(gProcessNameOffset - ACTIVELINK_OFFSETFROM_EPROC)		
#define PROCID_OFFSETFROM_ACTIVELINK		(-4)
#define EPROC_OFFSETFROM_WAITLISTENTRY		(0x1D0)
#define CID_OFFSETFROM_WAITLISTENTRY		(0x184)


#define GETPID(a) (*(DWORD *)((DWORD)(a) + PROCID_OFFSETFROM_ACTIVELINK))
#define GETPROCNAME(a) ((char *)(a) + PROCNAME_OFFSETFROM_ACTIVELINK)


//*********************************************************************************************
// Exported symbols that we need.
//
//*********************************************************************************************
extern PEPROCESS *PsInitialSystemProcess;

NTKERNELAPI NTSTATUS KeSetAffinityThread(long lParam1, long lParam2);

NTKERNELAPI PVOID MmGetSystemRoutineAddress(IN PUNICODE_STRING  SystemRoutineName); 

NTKERNELAPI NTSTATUS KeWaitForSingleObject(
  IN PVOID  Object,
  IN KWAIT_REASON  WaitReason,
  IN KPROCESSOR_MODE  WaitMode,
  IN BOOLEAN  Alertable,
  IN PLARGE_INTEGER Timeout OPTIONAL);

NTKERNELAPI NTSTATUS KeAddSystemServiceTable( 
        PVOID  lpAddressTable, 
        BOOLEAN  bUnknown,
        DWORD  dwNumEntries, 
        PVOID  lpParameterTable,
        DWORD dwTableID );

__declspec(dllimport) SDT KeServiceDescriptorTable;



//*********************************************************************************************
// Counts the number of processes by traversing ActiveProcessLink
//
//*********************************************************************************************

DWORD countProc(void)
{
    PEPROCESS curProc;
    LIST_ENTRY *entry, *start;
    DWORD count = 0;

	INTERRUPT_OFF
    curProc = *PsInitialSystemProcess;

    start = (LIST_ENTRY *)((DWORD)curProc + ACTIVELINK_OFFSETFROM_EPROC);
    entry = start;

    do
    {
		if(entry->Flink != start)		// skip PsActiveListHead
			count++;
        entry = entry->Flink;
    } while(entry != start);

	INTERRUPT_ON

	return count;
}


//*********************************************************************************************
// Gathers PID and ImageNames by traversing ActiveProcessLink
//
//*********************************************************************************************

NTSTATUS getProcList(char *outBuf, DWORD *size)
{
    PEPROCESS curProc;
    LIST_ENTRY *entry, *start;
    char *procName;
	DWORD numRec = *size / sizeof(PROCINFO);
	DWORD recCount = 0;
	PROCINFO *pPtr = (PROCINFO *)outBuf;

	if(numRec <= 0 || outBuf == NULL)
		return STATUS_INFO_LENGTH_MISMATCH;

	INTERRUPT_OFF
    
	curProc = *PsInitialSystemProcess;
	start = (LIST_ENTRY *)((DWORD)curProc + ACTIVELINK_OFFSETFROM_EPROC);		// offset 0xa0 into EPROCSS contains struct LIST_ENTRY ActiveProcessLinks

	entry = start;
	do
    {
		if(entry->Flink != start)		// skip PsActiveListHead
		{
			if(recCount >= numRec)
				break;

			pPtr->pId = GETPID(entry);
			RtlCopyMemory(pPtr->pName, GETPROCNAME(entry), NT_PROCNAMELEN);
			pPtr++;
			recCount++;			
		}
        entry = entry->Flink;

    } while(entry != start);

	INTERRUPT_ON

	if(recCount == numRec && entry != start)
		return STATUS_INFO_LENGTH_MISMATCH;

	*size = recCount * sizeof(PROCINFO);
	return STATUS_SUCCESS;
}


//*********************************************************************************************
// Counts the number of loaded modules from PsLoadedModuleList
//
//*********************************************************************************************

DWORD countMod(void)
{
	LIST_ENTRY *start, *cur;
	UNICODE_STRING *str;
	LIST_ENTRY *PsLoadedModuleList = (LIST_ENTRY *)gPsLoadedModuleList;
	DWORD count = 0;
	
	if(!gPsLoadedModuleList)
		return 0;

	cur = PsLoadedModuleList->Flink;
	do
	{
		count++;		
		cur = cur->Flink;
	} while(cur != PsLoadedModuleList);


	return count;
}


//*********************************************************************************************
// Gathers a list of all loaded modules from PsLoadedModuleList
//
//*********************************************************************************************

NTSTATUS getModList(char *outBuf, DWORD *size)
{
	LIST_ENTRY *start, *cur;
	ANSI_STRING astr;
	UNICODE_STRING *str;
	LIST_ENTRY *PsLoadedModuleList = (LIST_ENTRY *)gPsLoadedModuleList;
	DWORD numRec = *size / sizeof(DRIVERINFO);
	DWORD recCount = 0;
	DRIVERINFO *pPtr = (DRIVERINFO *)outBuf;
	
	if(!gPsLoadedModuleList)
		return STATUS_NOT_IMPLEMENTED;

	if(numRec <= 0 || outBuf == NULL)
		return STATUS_INFO_LENGTH_MISMATCH;


	cur = PsLoadedModuleList->Flink;
	do
	{
		if(recCount >= numRec)
			break;

		str = (UNICODE_STRING *)((DWORD)cur + 0x24);			// offset 0x24 - pointer to unicode string of drv name
				
		if(RtlUnicodeStringToAnsiString(&astr, str, TRUE) == STATUS_SUCCESS)	// sorry English windows only
		{
			strncpy(pPtr->drvName, astr.Buffer, MAX_PATH-1);
			RtlFreeAnsiString(&astr);

			pPtr->loadAddr = *(DWORD *)((DWORD)cur + 0x18);			// offset 0x18 - load address
			pPtr->imageSize = *(DWORD *)((DWORD)cur + 0x20);			// offset 0x18 - load address
			pPtr++;
			recCount++;
		}
		cur = cur->Flink;

	} while(cur != PsLoadedModuleList);

	
	if(recCount == numRec && cur != PsLoadedModuleList)
		return STATUS_INFO_LENGTH_MISMATCH;

	*size = recCount * sizeof(DRIVERINFO);

	return STATUS_SUCCESS;
}


//*********************************************************************************************
// Counts the current number of processes in ActiveProcessList
//
//*********************************************************************************************

DWORD countSchProc(void)
{
	if(gKiWaitInListHead == 0 || gKiWaitOutListHead == 0)
		return 0;
	
	return countProc();
}



//*********************************************************************************************
// Gathers PID and ImageNames from WaitInListHead and WaitOutListHead
//
//*********************************************************************************************

NTSTATUS getSchProcList(char *outBuf, DWORD *bufSize)
{
	KIRQL kq;
	LIST_ENTRY *cur, *start;
	DWORD val;
	PROCINFO *pid = NULL;
	DWORD count = 0, i;
	BOOL found = FALSE;
	DWORD numRec = *bufSize / sizeof(PROCINFO);
	NTSTATUS retVal = STATUS_SUCCESS;

	if(gKiWaitInListHead == 0 || gKiWaitOutListHead == 0)
		return STATUS_NOT_IMPLEMENTED;

	if(numRec == 0)
		return STATUS_INFO_LENGTH_MISMATCH;

	pid = (PROCINFO *)outBuf;
	if(!pid)
		return STATUS_INFO_LENGTH_MISMATCH;

	kq = KeRaiseIrqlToDpcLevel();
	INTERRUPT_OFF

	start = (LIST_ENTRY *)gKiWaitInListHead;
	cur = start->Flink;		

	do
	{
		found = FALSE;
		val = *((DWORD *)((DWORD)cur + EPROC_OFFSETFROM_WAITLISTENTRY));			// CID field offset 0x184 into ETHREAD.WaitListEntry
		for(i = 0; i < count && !found; i++)
		{
			if(pid[i].pId == val)
				found = TRUE;
		}
		if(!found)
		{
			// EPROCESS field offset 0x1D0 from ETHREAD.WaitListEntry
			char *namePtr = (char *)(*((DWORD *)((DWORD)cur + EPROC_OFFSETFROM_WAITLISTENTRY))) + gProcessNameOffset;		// proc name pos;			

			if(count < numRec)
			{
				pid[count].pId = val;
				RtlCopyMemory(pid[count].pName, namePtr, NT_PROCNAMELEN);
				count++;
			}
			else
			{
				retVal = STATUS_INFO_LENGTH_MISMATCH;
				goto done;
			}
		}

		cur = cur->Flink;

	} while(cur != start);


	start = (LIST_ENTRY *)gKiWaitOutListHead;
	cur = start->Flink;		

	do
	{
		found = FALSE;

		val = *((DWORD *)((DWORD)cur + EPROC_OFFSETFROM_WAITLISTENTRY));			// CID field offset 0x184 into ETHREAD.WaitListEntry
		for(i = 0; i < count && !found; i++)
		{
			if(pid[i].pId == val)
				found = TRUE;
		}
		if(!found)
		{
			// EPROCESS field offset 0x1D0 from ETHREAD.WaitListEntry
			char *namePtr = (char *)(*((DWORD *)((DWORD)cur + EPROC_OFFSETFROM_WAITLISTENTRY))) + gProcessNameOffset;		// proc name pos;			

			if(count < numRec)
			{
				pid[count].pId = val;
				RtlCopyMemory(pid[count].pName, namePtr, NT_PROCNAMELEN);
				count++;
			}
			else
			{
				retVal = STATUS_INFO_LENGTH_MISMATCH;
				goto done;
			}
		}

		cur = cur->Flink;

	} while(cur != start);


	if(count)
	{
		for(i = 0; i < count; i++)
		{
			pid[i].pId = *((DWORD *)(pid[i].pId + ACTIVELINK_OFFSETFROM_EPROC + PROCID_OFFSETFROM_ACTIVELINK));
		}
		*bufSize = count * sizeof(PROCINFO);
	}
done:
	INTERRUPT_ON
	KeLowerIrql(kq);

	return retVal;
}


//*********************************************************************************************
// Returns the name, load address and image size of a module
//
//*********************************************************************************************

BOOL getModuleInfo(char *modName, DWORD *loadAddr, DWORD *imageSize)
{
	LIST_ENTRY *start, *cur;
	UNICODE_STRING *str;
	ANSI_STRING aStr;
	LIST_ENTRY *PsLoadedModuleList = (LIST_ENTRY *)gPsLoadedModuleList;
	BOOL found = FALSE;
	
	if(!gPsLoadedModuleList)
		return FALSE;

	cur = PsLoadedModuleList->Flink;
	do
	{		
		str = (UNICODE_STRING *)((DWORD)cur + 0x24);
		if(RtlUnicodeStringToAnsiString(&aStr, str, TRUE) == STATUS_SUCCESS)
		{
			_strlwr(aStr.Buffer);
			if(strstr(aStr.Buffer, modName))
			{
				*loadAddr = *(DWORD *)((DWORD)cur + 0x18);
				*imageSize = *(DWORD *)((DWORD)cur + 0x20);
				found = TRUE;
				break;
			}
			RtlFreeAnsiString(&aStr);
		}
		
		cur = cur->Flink;
	} while(cur != PsLoadedModuleList);	

	return found;
}


//*********************************************************************************************
// Returns a list of addresses of the native APIs
//
//*********************************************************************************************

NTSTATUS findHookedService(char *outBuf, DWORD *size)
{
	DWORD loadAddr, imageSize;
	DWORD i;
	DWORD numRec = *size / sizeof(SDTINFO);
	SDTINFO *ptr;
	LPVOID modList;

	if(numRec < KeServiceDescriptorTable.ServiceLimit)
		return STATUS_INFO_LENGTH_MISMATCH;

	if(getModuleInfo("ntoskrnl.exe",&loadAddr, &imageSize))
	{
		DWORD limit = loadAddr + imageSize;
		ptr = (SDTINFO *)outBuf;

		for(i = 0; i < KeServiceDescriptorTable.ServiceLimit; i++)
		{
			ptr[i].sdtAddr = KeServiceDescriptorTable.ServiceTable[i];

			if(!(KeServiceDescriptorTable.ServiceTable[i] > loadAddr &&
				 KeServiceDescriptorTable.ServiceTable[i] < limit))
			{
				//DbgPrint("Entry %i <%X>\n", i, KeServiceDescriptorTable.ServiceTable[i]);
				ptr[i].isHooked = TRUE;
			}
			else
				ptr[i].isHooked = FALSE;
		}
	}
	else
		return STATUS_NOT_IMPLEMENTED;

	*size = KeServiceDescriptorTable.ServiceLimit * sizeof(SDTINFO);
	return STATUS_SUCCESS;
}


//*********************************************************************************************
// Returns a list of addresses of the GDI APIs
//
//*********************************************************************************************

NTSTATUS findHookedGDIService(char *outBuf, DWORD *size)
{
	DWORD loadAddr, imageSize;
	DWORD i;
	DWORD numRec = *size / sizeof(SDTINFO);
	SDTINFO *ptr;
	LPVOID modList;

	if(numRec < KeServiceDescriptorTableShadow[1].ServiceLimit)
		return STATUS_INFO_LENGTH_MISMATCH;

	if(getModuleInfo("win32k.sys", &loadAddr, &imageSize))
	{
		DWORD limit = loadAddr + imageSize;
		ptr = (SDTINFO *)outBuf;

		for(i = 0; i < KeServiceDescriptorTableShadow[1].ServiceLimit; i++)
		{
			ptr[i].sdtAddr = KeServiceDescriptorTableShadow[1].ServiceTable[i];

			if(!(KeServiceDescriptorTableShadow[1].ServiceTable[i] > loadAddr &&
				 KeServiceDescriptorTableShadow[1].ServiceTable[i] < limit))
			{
				//DbgPrint("Entry %i <%X>\n", i, KeServiceDescriptorTableShadow[1].ServiceTable[i]);
				ptr[i].isHooked = TRUE;
			}
			else
				ptr[i].isHooked = FALSE;
		}
	}
	else
		return STATUS_NOT_IMPLEMENTED;

	*size = KeServiceDescriptorTableShadow[1].ServiceLimit * sizeof(SDTINFO);
	return STATUS_SUCCESS;
}


//*********************************************************************************************
// Create our device to communicate with user land
//
//*********************************************************************************************

NTSTATUS createOurDevice(PDRIVER_OBJECT driverObject)
{
	PDEVICE_OBJECT pDeviceObject = NULL;
	NTSTATUS NTs;
	int i;

	NTs = IoCreateDevice(driverObject, 0, &gDeviceName,
					FILE_DEVICE_UNKNOWN, 0, FALSE, &pDeviceObject);

	if(NTs == STATUS_SUCCESS)
	{
		NTs = IoCreateSymbolicLink(&gSymbolicLinkName, &gDeviceName);

		if (NTs == STATUS_SUCCESS)
		{
			gpDeviceObject = pDeviceObject;

			DbgPrint("Device created successfully. Name = %S\n", gDeviceName.Buffer);
		}
		else
		{
			DbgPrint("Error while creating symbolic link\n");
            IoDeleteDevice (pDeviceObject);
		}

	}
	else
	{
		DbgPrint("Error while creating device\n");
	}

	return NTs;
}


//*********************************************************************************************
// DeviceDispatcher
//
//*********************************************************************************************

NTSTATUS DeviceDispatcher(PIRP pIrp)
{
    PIO_STACK_LOCATION esp;
    DWORD dInfo = 0;
	char *output;
    NTSTATUS ntS = STATUS_NOT_IMPLEMENTED;

	//DbgPrint("DeviceDispatcher called\n");
    esp = IoGetCurrentIrpStackLocation(pIrp);

    switch (esp->MajorFunction)
	{
		case IRP_MJ_CREATE:
			//DbgPrint("IRP_MJ_CREATE\n");
            ntS = STATUS_SUCCESS;
            break;

		case IRP_MJ_CLEANUP:
			//DbgPrint("IRP_MJ_CLEANUP\n");
            ntS = STATUS_SUCCESS;
            break;

		case IRP_MJ_CLOSE:
			//DbgPrint("IRP_MJ_CLOSE\n");
            ntS = STATUS_SUCCESS;
            break;

		case IRP_MJ_DEVICE_CONTROL:
			{
				DWORD sizeReq;

				// Returns list of installed drivers
				if(esp->Parameters.DeviceIoControl.IoControlCode == IOCTL_CODE1)
				{
					sizeReq = countMod() * sizeof(DRIVERINFO);
					if(sizeReq > 0)
					{
						if(esp->Parameters.DeviceIoControl.OutputBufferLength < sizeReq)
						{
							ntS = STATUS_INFO_LENGTH_MISMATCH;
						}
						else
						{
							LPVOID outBuf;
							DWORD outSize = esp->Parameters.DeviceIoControl.OutputBufferLength;
							//DbgPrint("inSize = %d. Good.\n", esp->Parameters.DeviceIoControl.OutputBufferLength);

							outBuf = MmGetSystemAddressForMdlSafe(pIrp->MdlAddress, NormalPagePriority);
							
							if(outBuf)
							{
								ntS = getModList((char *)outBuf, &outSize);
								if(ntS == STATUS_SUCCESS)
									dInfo = outSize;
							}
							else
								ntS = STATUS_SUCCESS;
						}
					}
				}

				// Returns list of processes from ActiveProcessLinks
				else if(esp->Parameters.DeviceIoControl.IoControlCode == IOCTL_CODE2)
				{
					sizeReq = countProc() * sizeof(PROCINFO);
					if(gProcessNameOffset && sizeReq > 0)
					{
						if(esp->Parameters.DeviceIoControl.OutputBufferLength < sizeReq)
						{
							ntS = STATUS_INFO_LENGTH_MISMATCH;
						}
						else
						{
							LPVOID outBuf;
							DWORD outSize = esp->Parameters.DeviceIoControl.OutputBufferLength;
							//DbgPrint("inSize = %d. Good.\n", esp->Parameters.DeviceIoControl.OutputBufferLength);

							outBuf = MmGetSystemAddressForMdlSafe(pIrp->MdlAddress, NormalPagePriority);
							
							if(outBuf)
							{
								ntS = getProcList((char *)outBuf, &outSize); 
								if(ntS == STATUS_SUCCESS)
									dInfo = outSize;
							}
							else
								ntS = STATUS_SUCCESS;
						}
					}
				}

				// Returns list of processes gathered from KiWaitInList and KiWaitOutList
				else if(esp->Parameters.DeviceIoControl.IoControlCode == IOCTL_CODE3)
				{
					sizeReq = countSchProc() * sizeof(PROCINFO);
					if(gProcessNameOffset && sizeReq > 0)
					{
						if(esp->Parameters.DeviceIoControl.OutputBufferLength < sizeReq)
						{
							ntS = STATUS_INFO_LENGTH_MISMATCH;
						}
						else
						{
							LPVOID outBuf;
							DWORD outSize = esp->Parameters.DeviceIoControl.OutputBufferLength;
							//DbgPrint("inSize = %d. Good.\n", esp->Parameters.DeviceIoControl.OutputBufferLength);

							outBuf = MmGetSystemAddressForMdlSafe(pIrp->MdlAddress, NormalPagePriority);
							
							if(outBuf)
							{
								ntS = getSchProcList((char *)outBuf, &outSize);

								if(ntS == STATUS_SUCCESS) 
									dInfo = outSize;
							}
							else
							{
								// cannot get double mapped memory
								ntS = STATUS_SUCCESS;
							}
						}
					}
				}

				// returns entries in the ntoskrnl service table
				else if(esp->Parameters.DeviceIoControl.IoControlCode == IOCTL_CODE4)
				{
					sizeReq = KeServiceDescriptorTable.ServiceLimit * sizeof(SDTINFO);
					if(sizeReq > 0)
					{
						if(esp->Parameters.DeviceIoControl.OutputBufferLength < sizeReq)
							ntS = STATUS_INFO_LENGTH_MISMATCH;
						else
						{
							LPVOID outBuf;
							DWORD outSize = esp->Parameters.DeviceIoControl.OutputBufferLength;	
							outBuf = MmGetSystemAddressForMdlSafe(pIrp->MdlAddress, NormalPagePriority);
							
							if(outBuf)
							{
								ntS = findHookedService((char *)outBuf, &outSize);

								if(ntS == STATUS_SUCCESS) 
									dInfo = outSize;
							}
							else	// cannot get double mapped memory								
								ntS = STATUS_SUCCESS;
						}
					}
				}

				// returns entries in the GDI service table
				else if(esp->Parameters.DeviceIoControl.IoControlCode == IOCTL_CODE5)
				{
					if(KeServiceDescriptorTableShadow != NULL)
					{
						sizeReq = KeServiceDescriptorTableShadow[1].ServiceLimit * sizeof(SDTINFO);
						if(sizeReq > 0)
						{
							if(esp->Parameters.DeviceIoControl.OutputBufferLength < sizeReq)
								ntS = STATUS_INFO_LENGTH_MISMATCH;
							else
							{
								LPVOID outBuf;
								DWORD outSize = esp->Parameters.DeviceIoControl.OutputBufferLength;	
								outBuf = MmGetSystemAddressForMdlSafe(pIrp->MdlAddress, NormalPagePriority);
								
								if(outBuf)
								{
									ntS = findHookedGDIService((char *)outBuf, &outSize);

									if(ntS == STATUS_SUCCESS) 
										dInfo = outSize;
								}
								else	// cannot get double mapped memory								
									ntS = STATUS_SUCCESS;
							}
						}
					}
				}

				// returns address of KeServiceDescriptorTable and KeServiceDescriptorTableShadow
				else if(esp->Parameters.DeviceIoControl.IoControlCode == IOCTL_CODE6)
				{
					if(esp->Parameters.DeviceIoControl.OutputBufferLength >= 16)
					{
						DWORD *outDword = (DWORD *)pIrp->AssociatedIrp.SystemBuffer;
						if(outDword)
						{
							outDword[0] = (DWORD)&KeServiceDescriptorTable;
							outDword[1] = (DWORD)(KeServiceDescriptorTable.ServiceTable);
							outDword[2] = (DWORD)KeServiceDescriptorTableShadow;
							outDword[3] = (DWORD)(KeServiceDescriptorTableShadow[1].ServiceTable);
							dInfo = 16;
							ntS = STATUS_SUCCESS;
						}
					}
					else
						ntS = STATUS_INFO_LENGTH_MISMATCH;
				}

				break;
			}

		default:
		    break;
	}

    pIrp->IoStatus.Status      = ntS;
    pIrp->IoStatus.Information = dInfo;
    IoCompleteRequest(pIrp, IO_NO_INCREMENT);

    return(ntS);
}



//*********************************************************************************************
// DriverDispatcher
//
//*********************************************************************************************

NTSTATUS DriverDispatcher(PDEVICE_OBJECT pDeviceObject, PIRP pIrp)
{
	NTSTATUS ntS = STATUS_INVALID_PARAMETER_1;

	//DbgPrint("DriverDispatcher called\n");

	if (pDeviceObject == gpDeviceObject)
		ntS = DeviceDispatcher(pIrp);

    return(ntS);
}



//*********************************************************************************************
// DriverUnload - removes our device object
//
//*********************************************************************************************

VOID DriverUnload(IN PDRIVER_OBJECT DriverObject)
{
	DbgPrint("Unloading...\n");

	IoDeleteSymbolicLink(&gSymbolicLinkName);

	if(gpDeviceObject)
		IoDeleteDevice(gpDeviceObject);
}


//*********************************************************************************************
// Attempt to locate the non-exported symbol PsLoadedModuleList by
// scanning MmGetSystemRoutineAddress for
//
// 8b35b8e14680     mov     esi,[nt!PsLoadedModuleList (8046e1b8)]
// 81feb8e14680     cmp     esi,0x8046e1b8
//
// Tested on Win2k SP2 and SP4
//*********************************************************************************************

DWORD findLoadedModuleList(void)
{
	int i;
	unsigned char *ptr;
	DWORD retVal = 0, tempVal;
	ptr = (unsigned char *)MmGetSystemRoutineAddress;
	
    for( i = 0; i < PAGE_SIZE; i++ ) 
	{
        if(*ptr == 0x8b && *(ptr+1) == 0x35 && *(ptr+6) == 0x81 && *(ptr+7) == 0xfe)
		{
            tempVal = *(DWORD *)(ptr + 2);
			if(tempVal == *(DWORD *)(ptr + 8))
			{
				retVal = tempVal;
				break;
			}
		}
		ptr++;
    }
	return retVal;
}


//*********************************************************************************************
// Attempt to locate the non-exported symbols KiWaitInListHead and KiWaitOutListHead by
// scanning in KeWaitForSingleObject API for
//
// mov     ecx,0x80482258   <- KiWaitInListHead
// test    al,al
//
// And,
// mov     ecx, offset dword_0_482808  <- KiWaitOutListHead
//
// Tested on Win2k SP2 and SP4
//*********************************************************************************************

void findKiWaitInOutListHead(DWORD *waitInListAddr, DWORD *waitOutListAddr)
{
	int i;
	unsigned char *ptr;
	DWORD retVal1 = 0, retVal2 = 0;
	ptr = (unsigned char *)KeWaitForSingleObject;
	
    for( i = 0; i < PAGE_SIZE; i++ ) 
	{
        if(*ptr == 0xb9 && *(ptr+5) == 0x84 && *(ptr+6) == 0xc0 && *(ptr+24) == 0xb9)
		{
            retVal1 = *(DWORD *)(ptr + 1);
			retVal2 = *(DWORD *)(ptr + 25);

			break;
		}
		ptr++;
    }

	*waitInListAddr = retVal1;
	*waitOutListAddr = retVal2;

	return;
}


//*********************************************************************************************
// Attempt to locate the non-exported symbol KiDispatcherReadyListHead by
// scanning KeSetAffinityThread for
//
// 8d04cde0224880	lea eax,[nt!KiDispatcherReadyListHead (804822e0)+ecx*8]
// 3900				cmp     [eax],eax
//
// Tested on Win2k SP2 and SP4
//*********************************************************************************************

DWORD findKiDispatcherReadyListHead(void)
{
	int i;
	unsigned char *ptr;
	DWORD retVal = 0;
	ptr = (unsigned char *)KeSetAffinityThread;
	
    for( i = 0; i < PAGE_SIZE; i++ ) 
	{
        if(*ptr == 0x8d && *(ptr+1) == 0x04 && *(ptr+2) == 0xcd && *(ptr+7) == 0x39)
		{
            retVal = *(DWORD *)(ptr + 3);
			break;
		}
		ptr++;
    }
	return retVal;
}


//*********************************************************************************************
// Gets the offset to process name, thx to SysInternals for this idea.
// This will not work if driver is loaded using SystemLoadAndCallImage
// 
//*********************************************************************************************

DWORD getProcNameOffset(void)
{
    PEPROCESS curproc;
    int i;
	DWORD procNameOffset = 0;

	curproc = PsGetCurrentProcess();

    for(i = 0; i < PAGE_SIZE; i++)
	{
        if(strncmp( "System", (PCHAR) curproc + i, strlen("System")) == 0)
		{
            procNameOffset = i;
			break;
		}
    }

	return procNameOffset;
}


//*********************************************************************************************
// Find address of KeServiceDescriptorTableShadow by scanning KeAddSystemServiceTable
//
//*********************************************************************************************

DWORD findAddressofShadowTable()
{
	int i;
	unsigned char *p;
	DWORD val;

	p = (unsigned char *)KeAddSystemServiceTable;

	for (i = 0; i < PAGE_SIZE; i++, p++)
	{
		__try
		{
			val = *(unsigned int *)p;
		}
		__except (EXCEPTION_EXECUTE_HANDLER)
		{
			return 0;
		}

		if (MmIsAddressValid((PVOID)val))
		{
			if (memcmp((PVOID)val, &KeServiceDescriptorTable, 16) == 0)
			{
				if((PVOID)val != &KeServiceDescriptorTable)
					return val;
			}
		}
	}
	return 0;
}


//*********************************************************************************************
// DriverEntry
//
//*********************************************************************************************

NTSTATUS DriverEntry(PDRIVER_OBJECT driverObject, PUNICODE_STRING RegistryPath)
{
    int i;
	NTSTATUS ns;
	
	KIRQL kq;

    driverObject->DriverUnload = DriverUnload;

	// finds address of non-exported symbols
	gPsLoadedModuleList = findLoadedModuleList();
	findKiWaitInOutListHead(&gKiWaitInListHead, &gKiWaitOutListHead);
	gKiDispatcherReadyListHead = findKiDispatcherReadyListHead();
	gProcessNameOffset = getProcNameOffset();
	KeServiceDescriptorTableShadow = (SDT *)findAddressofShadowTable();
	
 	for (i = 0; i < IRP_MJ_MAXIMUM_FUNCTION; i++)
	{
   		driverObject->MajorFunction[i] = DriverDispatcher;
	}

	createOurDevice(driverObject);
	
	DbgPrint("PsLoadedModuleList %X\n", gPsLoadedModuleList);
	DbgPrint("KiWaitInListHead %X\n", gKiWaitInListHead);
	DbgPrint("KiWaitOutListHead %X\n", gKiWaitOutListHead);
	DbgPrint("KiDispatcherReadyListHead %X\n", gKiDispatcherReadyListHead);
	DbgPrint("ProcessNameOffset %X\n", gProcessNameOffset);
	DbgPrint("KeServiceDescriptorTableShadow %X\n", KeServiceDescriptorTableShadow);

    return(STATUS_SUCCESS);
}
